"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.VectorClock = exports.LogicalClock = exports.LogicalTimespan = exports.LogicalTimestamp = void 0;
class LogicalTimestamp {
    constructor(sessionId, time) {
        this.sessionId = sessionId;
        this.time = time;
    }
    isEqual(ts) {
        return (this.sessionId === ts.sessionId) && (this.time === ts.time);
    }
    compare(ts) {
        if (this.time > ts.time)
            return 1;
        if (this.time < ts.time)
            return -1;
        if (this.sessionId > ts.sessionId)
            return 1;
        if (this.sessionId < ts.sessionId)
            return -1;
        return 0;
    }
    inSpan(span, ts, tsSpan) {
        if (this.sessionId !== ts.sessionId)
            return false;
        if (this.time > ts.time)
            return false;
        if ((this.time + span) < (ts.time + tsSpan))
            return false;
        return true;
    }
    overlap(span, ts, tsSpan) {
        if (this.sessionId !== ts.sessionId)
            return 0;
        const x1 = this.time;
        const x2 = x1 + span;
        const y1 = ts.time;
        const y2 = y1 + tsSpan;
        const min = Math.max(x1, y1);
        const max = Math.min(x2, y2);
        const diff = max - min;
        return diff <= 0 ? 0 : diff;
    }
    tick(cycles) {
        return new LogicalTimestamp(this.sessionId, this.time + cycles);
    }
    interval(cycles, span) {
        return new LogicalTimespan(this.sessionId, this.time + cycles, span);
    }
    toString() {
        return this.sessionId + '!' + this.time;
    }
    toDisplayString() {
        let session = String(this.sessionId);
        if (session.length > 4)
            session = '..' + session.substr(session.length - 4);
        return session + '!' + this.time;
    }
    clock() {
        return new LogicalClock(this.sessionId, this.time);
    }
    compact() {
        return this.sessionId + ',' + this.time;
    }
}
exports.LogicalTimestamp = LogicalTimestamp;
class LogicalTimespan extends LogicalTimestamp {
    constructor(sessionId, time, span) {
        super(sessionId, time);
        this.span = span;
    }
}
exports.LogicalTimespan = LogicalTimespan;
class LogicalClock extends LogicalTimestamp {
    constructor(sessionId, time) {
        super(sessionId, 0);
        this.sessionId = sessionId;
        this.time = time;
    }
    tick(cycles) {
        const timestamp = new LogicalTimestamp(this.sessionId, this.time);
        this.time += cycles;
        return timestamp;
    }
}
exports.LogicalClock = LogicalClock;
class VectorClock extends LogicalClock {
    constructor(sessionId, time) {
        super(sessionId, time);
        this.clocks = new Map();
        this.clocks.set(sessionId, this);
    }
    observe(ts, span) {
        const time = ts.time + span - 1;
        const clock = this.clocks.get(ts.sessionId);
        if (!clock)
            this.clocks.set(ts.sessionId, ts.tick(span - 1));
        else if (time > clock.time)
            clock.time = ts.time;
        if (time >= this.time)
            this.time = time + 1;
    }
    serialize() {
        let str = '[' + this.sessionId + ',' + this.time;
        for (const clock of this.clocks.values())
            if (clock.sessionId !== this.sessionId)
                str += ',' + clock.sessionId + ',' + clock.time;
        return str + ']';
    }
    static deserialize(data) {
        const [sessionId, time] = data;
        const clock = new VectorClock(sessionId, time);
        const length = data.length;
        let i = 2;
        while (i < length) {
            const sessionId = data[i++];
            const time = data[i++];
            clock.observe(new LogicalTimestamp(sessionId, time), 1);
        }
        return clock;
    }
    fork(sessionId) {
        const clock = new VectorClock(sessionId, this.time);
        for (const ts of this.clocks.values())
            clock.observe(ts.tick(0), 1);
        return clock;
    }
    clone() {
        return this.fork(this.sessionId);
    }
}
exports.VectorClock = VectorClock;
