"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Document = void 0;
const constants_1 = require("./constants");
const IdentifiableIndex_1 = require("./IdentifiableIndex");
const util_1 = require("./util");
const SetRootOperation_1 = require("../json-crdt-patch/operations/SetRootOperation");
const clock_1 = require("../json-crdt-patch/clock");
const DocRootType_1 = require("./types/lww-doc-root/DocRootType");
const ObjectType_1 = require("./types/lww-object/ObjectType");
const ArrayType_1 = require("./types/rga-array/ArrayType");
const StringType_1 = require("./types/rga-string/StringType");
const MakeObjectOperation_1 = require("../json-crdt-patch/operations/MakeObjectOperation");
const SetObjectKeysOperation_1 = require("../json-crdt-patch/operations/SetObjectKeysOperation");
const MakeArrayOperation_1 = require("../json-crdt-patch/operations/MakeArrayOperation");
const InsertArrayElementsOperation_1 = require("../json-crdt-patch/operations/InsertArrayElementsOperation");
const constants_2 = require("../json-crdt-patch/constants");
const DeleteOperation_1 = require("../json-crdt-patch/operations/DeleteOperation");
const MakeStringOperation_1 = require("../json-crdt-patch/operations/MakeStringOperation");
const InsertStringSubstringOperation_1 = require("../json-crdt-patch/operations/InsertStringSubstringOperation");
const JsonPatch_1 = require("./JsonPatch");
const json_patch_1 = require("../json-patch");
const DocumentApi_1 = require("./api/DocumentApi");
const MakeValueOperation_1 = require("../json-crdt-patch/operations/MakeValueOperation");
const ValueType_1 = require("./types/lww-value/ValueType");
const SetValueOperation_1 = require("../json-crdt-patch/operations/SetValueOperation");
class Document {
    constructor(clock = new clock_1.VectorClock(util_1.random40BitInt(), 0)) {
        this.root = new DocRootType_1.DocRootType(this, constants_2.ORIGIN, null);
        this.nodes = new IdentifiableIndex_1.IdentifiableIndex();
        this.clock = clock;
        this.nodes.index(constants_1.NULL);
        this.nodes.index(constants_1.TRUE);
        this.nodes.index(constants_1.FALSE);
        this.nodes.index(constants_1.UNDEFINED);
        this.api = new DocumentApi_1.DocumentApi(this);
    }
    applyPatch(patch) {
        const ops = patch.ops;
        const { length } = ops;
        for (let i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    }
    applyOperation(op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation_1.MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType_1.ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation_1.MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType_1.ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation_1.MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType_1.StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation_1.MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType_1.ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation_1.SetRootOperation) {
            const oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation_1.SetObjectKeysOperation) {
            const obj = this.nodes.get(op.object);
            if (!(obj instanceof ObjectType_1.ObjectType))
                return;
            obj.insert(op);
        }
        else if (op instanceof SetValueOperation_1.SetValueOperation) {
            const obj = this.nodes.get(op.obj);
            if (!(obj instanceof ValueType_1.ValueType))
                return;
            obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation_1.InsertArrayElementsOperation) {
            const arr = this.nodes.get(op.arr);
            if (!(arr instanceof ArrayType_1.ArrayType))
                return;
            arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation_1.InsertStringSubstringOperation) {
            const arr = this.nodes.get(op.obj);
            if (!(arr instanceof StringType_1.StringType))
                return;
            arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation_1.DeleteOperation) {
            const node = this.nodes.get(op.obj);
            if (node instanceof ArrayType_1.ArrayType)
                node.delete(op);
            else if (node instanceof StringType_1.StringType)
                node.onDelete(op);
        }
    }
    deleteNodeTree(value) {
        const isSystemNode = value.sessionId < 1;
        if (isSystemNode)
            return;
        const node = this.nodes.get(value);
        if (!node)
            return;
        for (const child of node.children())
            this.deleteNodeTree(child);
        this.nodes.delete(value);
    }
    toJson() {
        return this.root.toJson();
    }
    clone() {
        const doc = new Document(this.clock.clone());
        doc.root = this.root.clone(doc);
        return doc;
    }
    fork() {
        const sessionId = util_1.random40BitInt();
        const doc = new Document(this.clock.fork(sessionId));
        doc.root = this.root.clone(doc);
        return doc;
    }
    toString() {
        const value = this.root.toValue();
        const op = this.nodes.get(value);
        return op ? op.toString('') : 'undefined';
    }
    find(steps) {
        const id = this.root.toValue();
        let node = this.nodes.get(id) || constants_1.NULL;
        const length = steps.length;
        if (!length)
            return node;
        let i = 0;
        while (i < length) {
            const step = steps[i++];
            if (node instanceof ObjectType_1.ObjectType) {
                const id = node.get(String(step));
                if (!id)
                    return constants_1.UNDEFINED;
                const nextNode = this.nodes.get(id);
                if (!nextNode)
                    return constants_1.UNDEFINED;
                node = nextNode;
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                const id = node.findValue(Number(step));
                const nextNode = this.nodes.get(id);
                if (!nextNode)
                    return constants_1.UNDEFINED;
                node = nextNode;
                continue;
            }
        }
        return node;
    }
    applyJsonPatch(operations) {
        const ops = operations.map(json_patch_1.operationToOp);
        const jsonPatch = new JsonPatch_1.JsonPatch(this);
        const draft = jsonPatch.fromOps(ops);
        const patch = draft.patch(this.clock);
        this.clock.tick(patch.span());
        this.applyPatch(patch);
    }
}
exports.Document = Document;
