"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.VectorClock = exports.LogicalClock = exports.LogicalTimespan = exports.LogicalTimestamp = void 0;
var tslib_1 = require("tslib");
var LogicalTimestamp = (function () {
    function LogicalTimestamp(sessionId, time) {
        this.sessionId = sessionId;
        this.time = time;
    }
    LogicalTimestamp.prototype.isEqual = function (ts) {
        return (this.sessionId === ts.sessionId) && (this.time === ts.time);
    };
    LogicalTimestamp.prototype.compare = function (ts) {
        if (this.time > ts.time)
            return 1;
        if (this.time < ts.time)
            return -1;
        if (this.sessionId > ts.sessionId)
            return 1;
        if (this.sessionId < ts.sessionId)
            return -1;
        return 0;
    };
    LogicalTimestamp.prototype.inSpan = function (span, ts, tsSpan) {
        if (this.sessionId !== ts.sessionId)
            return false;
        if (this.time > ts.time)
            return false;
        if ((this.time + span) < (ts.time + tsSpan))
            return false;
        return true;
    };
    LogicalTimestamp.prototype.overlap = function (span, ts, tsSpan) {
        if (this.sessionId !== ts.sessionId)
            return 0;
        var x1 = this.time;
        var x2 = x1 + span;
        var y1 = ts.time;
        var y2 = y1 + tsSpan;
        var min = Math.max(x1, y1);
        var max = Math.min(x2, y2);
        var diff = max - min;
        return diff <= 0 ? 0 : diff;
    };
    LogicalTimestamp.prototype.tick = function (cycles) {
        return new LogicalTimestamp(this.sessionId, this.time + cycles);
    };
    LogicalTimestamp.prototype.interval = function (cycles, span) {
        return new LogicalTimespan(this.sessionId, this.time + cycles, span);
    };
    LogicalTimestamp.prototype.toString = function () {
        return this.sessionId + '!' + this.time;
    };
    LogicalTimestamp.prototype.toDisplayString = function () {
        var session = String(this.sessionId);
        if (session.length > 4)
            session = '..' + session.substr(session.length - 4);
        return session + '!' + this.time;
    };
    LogicalTimestamp.prototype.clock = function () {
        return new LogicalClock(this.sessionId, this.time);
    };
    LogicalTimestamp.prototype.compact = function () {
        return this.sessionId + ',' + this.time;
    };
    return LogicalTimestamp;
}());
exports.LogicalTimestamp = LogicalTimestamp;
var LogicalTimespan = (function (_super) {
    tslib_1.__extends(LogicalTimespan, _super);
    function LogicalTimespan(sessionId, time, span) {
        var _this = _super.call(this, sessionId, time) || this;
        _this.span = span;
        return _this;
    }
    return LogicalTimespan;
}(LogicalTimestamp));
exports.LogicalTimespan = LogicalTimespan;
var LogicalClock = (function (_super) {
    tslib_1.__extends(LogicalClock, _super);
    function LogicalClock(sessionId, time) {
        var _this = _super.call(this, sessionId, 0) || this;
        _this.sessionId = sessionId;
        _this.time = time;
        return _this;
    }
    LogicalClock.prototype.tick = function (cycles) {
        var timestamp = new LogicalTimestamp(this.sessionId, this.time);
        this.time += cycles;
        return timestamp;
    };
    return LogicalClock;
}(LogicalTimestamp));
exports.LogicalClock = LogicalClock;
var VectorClock = (function (_super) {
    tslib_1.__extends(VectorClock, _super);
    function VectorClock(sessionId, time) {
        var _this = _super.call(this, sessionId, time) || this;
        _this.clocks = new Map();
        _this.clocks.set(sessionId, _this);
        return _this;
    }
    VectorClock.prototype.observe = function (ts, span) {
        var time = ts.time + span - 1;
        var clock = this.clocks.get(ts.sessionId);
        if (!clock)
            this.clocks.set(ts.sessionId, ts.tick(span - 1));
        else if (time > clock.time)
            clock.time = ts.time;
        if (time >= this.time)
            this.time = time + 1;
    };
    VectorClock.prototype.serialize = function () {
        var e_1, _a;
        var str = '[' + this.sessionId + ',' + this.time;
        try {
            for (var _b = tslib_1.__values(this.clocks.values()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var clock = _c.value;
                if (clock.sessionId !== this.sessionId)
                    str += ',' + clock.sessionId + ',' + clock.time;
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return str + ']';
    };
    VectorClock.deserialize = function (data) {
        var _a = tslib_1.__read(data, 2), sessionId = _a[0], time = _a[1];
        var clock = new VectorClock(sessionId, time);
        var length = data.length;
        var i = 2;
        while (i < length) {
            var sessionId_1 = data[i++];
            var time_1 = data[i++];
            clock.observe(new LogicalTimestamp(sessionId_1, time_1), 1);
        }
        return clock;
    };
    VectorClock.prototype.fork = function (sessionId) {
        var e_2, _a;
        var clock = new VectorClock(sessionId, this.time);
        try {
            for (var _b = tslib_1.__values(this.clocks.values()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var ts = _c.value;
                clock.observe(ts.tick(0), 1);
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_2) throw e_2.error; }
        }
        return clock;
    };
    VectorClock.prototype.clone = function () {
        return this.fork(this.sessionId);
    };
    return VectorClock;
}(LogicalClock));
exports.VectorClock = VectorClock;
