"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Document = void 0;
var tslib_1 = require("tslib");
var constants_1 = require("./constants");
var IdentifiableIndex_1 = require("./IdentifiableIndex");
var util_1 = require("./util");
var SetRootOperation_1 = require("../json-crdt-patch/operations/SetRootOperation");
var clock_1 = require("../json-crdt-patch/clock");
var DocRootType_1 = require("./types/lww-doc-root/DocRootType");
var ObjectType_1 = require("./types/lww-object/ObjectType");
var ArrayType_1 = require("./types/rga-array/ArrayType");
var StringType_1 = require("./types/rga-string/StringType");
var MakeObjectOperation_1 = require("../json-crdt-patch/operations/MakeObjectOperation");
var SetObjectKeysOperation_1 = require("../json-crdt-patch/operations/SetObjectKeysOperation");
var MakeArrayOperation_1 = require("../json-crdt-patch/operations/MakeArrayOperation");
var InsertArrayElementsOperation_1 = require("../json-crdt-patch/operations/InsertArrayElementsOperation");
var constants_2 = require("../json-crdt-patch/constants");
var DeleteOperation_1 = require("../json-crdt-patch/operations/DeleteOperation");
var MakeStringOperation_1 = require("../json-crdt-patch/operations/MakeStringOperation");
var InsertStringSubstringOperation_1 = require("../json-crdt-patch/operations/InsertStringSubstringOperation");
var JsonPatch_1 = require("./JsonPatch");
var json_patch_1 = require("../json-patch");
var DocumentApi_1 = require("./api/DocumentApi");
var MakeValueOperation_1 = require("../json-crdt-patch/operations/MakeValueOperation");
var ValueType_1 = require("./types/lww-value/ValueType");
var SetValueOperation_1 = require("../json-crdt-patch/operations/SetValueOperation");
var Document = (function () {
    function Document(clock) {
        if (clock === void 0) { clock = new clock_1.VectorClock(util_1.random40BitInt(), 0); }
        this.root = new DocRootType_1.DocRootType(this, constants_2.ORIGIN, null);
        this.nodes = new IdentifiableIndex_1.IdentifiableIndex();
        this.clock = clock;
        this.nodes.index(constants_1.NULL);
        this.nodes.index(constants_1.TRUE);
        this.nodes.index(constants_1.FALSE);
        this.nodes.index(constants_1.UNDEFINED);
        this.api = new DocumentApi_1.DocumentApi(this);
    }
    Document.prototype.applyPatch = function (patch) {
        var ops = patch.ops;
        var length = ops.length;
        for (var i = 0; i < length; i++)
            this.applyOperation(ops[i]);
    };
    Document.prototype.applyOperation = function (op) {
        this.clock.observe(op.id, op.span());
        if (op instanceof MakeObjectOperation_1.MakeObjectOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ObjectType_1.ObjectType(this, op.id));
        }
        else if (op instanceof MakeArrayOperation_1.MakeArrayOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ArrayType_1.ArrayType(this, op.id));
        }
        else if (op instanceof MakeStringOperation_1.MakeStringOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new StringType_1.StringType(this, op.id));
        }
        else if (op instanceof MakeValueOperation_1.MakeValueOperation) {
            if (!this.nodes.get(op.id))
                this.nodes.index(new ValueType_1.ValueType(op.id, op.id, op.value));
        }
        else if (op instanceof SetRootOperation_1.SetRootOperation) {
            var oldValue = this.root.insert(op);
            if (oldValue)
                this.deleteNodeTree(oldValue);
        }
        else if (op instanceof SetObjectKeysOperation_1.SetObjectKeysOperation) {
            var obj = this.nodes.get(op.object);
            if (!(obj instanceof ObjectType_1.ObjectType))
                return;
            obj.insert(op);
        }
        else if (op instanceof SetValueOperation_1.SetValueOperation) {
            var obj = this.nodes.get(op.obj);
            if (!(obj instanceof ValueType_1.ValueType))
                return;
            obj.insert(op);
        }
        else if (op instanceof InsertArrayElementsOperation_1.InsertArrayElementsOperation) {
            var arr = this.nodes.get(op.arr);
            if (!(arr instanceof ArrayType_1.ArrayType))
                return;
            arr.insert(op);
        }
        else if (op instanceof InsertStringSubstringOperation_1.InsertStringSubstringOperation) {
            var arr = this.nodes.get(op.obj);
            if (!(arr instanceof StringType_1.StringType))
                return;
            arr.onInsert(op);
        }
        else if (op instanceof DeleteOperation_1.DeleteOperation) {
            var node = this.nodes.get(op.obj);
            if (node instanceof ArrayType_1.ArrayType)
                node.delete(op);
            else if (node instanceof StringType_1.StringType)
                node.onDelete(op);
        }
    };
    Document.prototype.deleteNodeTree = function (value) {
        var e_1, _a;
        var isSystemNode = value.sessionId < 1;
        if (isSystemNode)
            return;
        var node = this.nodes.get(value);
        if (!node)
            return;
        try {
            for (var _b = tslib_1.__values(node.children()), _c = _b.next(); !_c.done; _c = _b.next()) {
                var child = _c.value;
                this.deleteNodeTree(child);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        this.nodes.delete(value);
    };
    Document.prototype.toJson = function () {
        return this.root.toJson();
    };
    Document.prototype.clone = function () {
        var doc = new Document(this.clock.clone());
        doc.root = this.root.clone(doc);
        return doc;
    };
    Document.prototype.fork = function () {
        var sessionId = util_1.random40BitInt();
        var doc = new Document(this.clock.fork(sessionId));
        doc.root = this.root.clone(doc);
        return doc;
    };
    Document.prototype.toString = function () {
        var value = this.root.toValue();
        var op = this.nodes.get(value);
        return op ? op.toString('') : 'undefined';
    };
    Document.prototype.find = function (steps) {
        var id = this.root.toValue();
        var node = this.nodes.get(id) || constants_1.NULL;
        var length = steps.length;
        if (!length)
            return node;
        var i = 0;
        while (i < length) {
            var step = steps[i++];
            if (node instanceof ObjectType_1.ObjectType) {
                var id_1 = node.get(String(step));
                if (!id_1)
                    return constants_1.UNDEFINED;
                var nextNode = this.nodes.get(id_1);
                if (!nextNode)
                    return constants_1.UNDEFINED;
                node = nextNode;
            }
            else if (node instanceof ArrayType_1.ArrayType) {
                var id_2 = node.findValue(Number(step));
                var nextNode = this.nodes.get(id_2);
                if (!nextNode)
                    return constants_1.UNDEFINED;
                node = nextNode;
                continue;
            }
        }
        return node;
    };
    Document.prototype.applyJsonPatch = function (operations) {
        var ops = operations.map(json_patch_1.operationToOp);
        var jsonPatch = new JsonPatch_1.JsonPatch(this);
        var draft = jsonPatch.fromOps(ops);
        var patch = draft.patch(this.clock);
        this.clock.tick(patch.span());
        this.applyPatch(patch);
    };
    return Document;
}());
exports.Document = Document;
